#include "stdafx.h"
#include "detectobjects.h"
#include "dump.h"
#include "integralhistogram.h"
#include "profiling.h"
#include "prune_subwindows.h"

#include <sstream>
#include <time.h>

using namespace std;

pair<CvMat*, CvMat*> ClassifyImagePixels(CvMat * source_image,
                                         CvMat * cluster_means_norm,
                                         PixClass pixclass)
{
    CvMat * timg = cvCreateMat(source_image->rows, source_image->cols, CV_16SC1);
    CvMat * wimg = cvCreateMat(source_image->rows, source_image->cols, CV_32FC1);
    ilClassifyImagePixels(source_image, timg, wimg, cluster_means_norm, (int)pixclass);
    return make_pair(timg, wimg);
}


CvMat ** IntegralBlockHist(CvMat * timg, CvMat * wimg,
                           size_t cluster_number,
                           size_t dx, size_t dy)
{
    size_t nby = (div(timg->rows, dy)).quot;
    size_t nbx = (div(timg->cols, dx)).quot;
    CvMat ** iblockhist = new CvMat*[cluster_number];
    for (size_t i = 0; i < cluster_number; i++)
        iblockhist[i] = cvCreateMat(nby, nbx, CV_32FC1);

//    ilIntegralBlockHist2( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("ilIntegralBlockHist2");

    ilIntegralBlockHist( timg, wimg, iblockhist, cluster_number, dx, dy);
//    Tick("IntegralBlockHist");
    return iblockhist;
}

IMPEXP
SubWindows DetectObjects( IplImage* src,
                          Model const & model, 
                          CvMat * cmeans,
                          CvMat * covmat,
                          CvMat * cmeansnorm,
						  int pixclass,
						  HistBlockSizeArray bsarray,
						  ScaleFctArray scarray)
{
    using namespace std;

    double dt;
    int64 t1,t2,t3,t01,t02;
	clock_t tc1,tc2;
	double totproctime=0, totdettime=0;

	//*** convert to one channel CvMat
    CvMat* srcgr = cvCreateMat(src->height,src->width,CV_8UC1);
	if (src->nChannels==1) cvCopy(src,srcgr);
	else cvCvtColor(src,srcgr,CV_RGB2GRAY);

	SubWindows subwindows;

	//*** iterate over scales
	for (ScaleFctArray::iterator sc = scarray.begin(); sc != scarray.end(); ++sc){

		double scf=*sc;
		int ysz=floor(scf*(src->height));
		int xsz=floor(scf*(src->width));

		if ((xsz>0) && (ysz>0)){
			//*** subsample image
			//printf("xsz: %d, ysz: %d\n",xsz,ysz);
			CvMat* srcimg = cvCreateMat(ysz,xsz,CV_8UC1);
			cvResize(srcgr,srcimg);
			CvMat *srcimg32f = cvCreateMat(ysz,xsz,CV_32FC1);
			cvConvert(srcimg,srcimg32f);	
	
			//********************************************************
			//***  MAIN PART FROM HERE....
			//***  COMPUTES HISTOGRAM OF IMAGE "srcimg"

			//*** label image pixels
			//int pixclass=TEXTONS_SC12_GDORDER2_;
			//int pixclass=ORIENTGRAD_DNUM4_;
			t01=cvGetTickCount();
			t1=cvGetTickCount();
			CvMat *timg = cvCreateMat(ysz,xsz,CV_16SC1);
			CvMat *wimg = cvCreateMat(ysz,xsz,CV_32FC1);

			tc1 = clock ();
			ilClassifyImagePixels(srcimg,timg,wimg,cmeansnorm,pixclass);
			if (0) { //DEBUG
			  cvSet(timg, cvScalar(1.0));
			  cvSet(wimg, cvScalar(1.0));
			}
			tc2 = clock ();
			totproctime+=(double)(tc2-tc1)/(double)CLK_TCK;
			//printf("\n  ClassifyImagePixels time:        %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);
			t2=cvGetTickCount();
			dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);	
			
			//*** iterate over block sizes
			for (HistBlockSizeArray::iterator bs = bsarray.begin(); bs != bsarray.end(); ++bs){	

				//*** compute integral weighted block histogram
				t1=cvGetTickCount();
				double dx=bs->dx, dy=bs->dy;
				int nc;

				if ((dx<xsz) && (dy<ysz)){
					//printf("  - dx:%1.1f, dy:%1.1f\n",dx,dy);
					if (pixclass==TEXTONS_SC12_GDORDER2_) nc=12;
					if (pixclass==ORIENTGRAD_DNUM4_) nc=4;
					int	nbx=floor((double)xsz/(double)dx);
					int nby=floor((double)ysz/(double)dy);
					CvMat **iblockhist = new CvMat*[nc];
					for (int i=0; i<nc; i++) iblockhist[i]=cvCreateMat(nby,nbx,CV_32FC1);
					// /*
					
					tc1 = clock ();
					if (pixclass==ORIENTGRAD_DNUM4_){
						ilIntegralBlockHist(timg,wimg,iblockhist,nc,dx,dy); // weighted hist
					}else{
						ilIntegralBlockHist(timg,NULL,iblockhist,nc,dx,dy); // unweighted hist
					}
					tc2 = clock ();
					totproctime+=(double)(tc2-tc1)/(double)CLK_TCK;
					//printf("  IntegralBlockHist time:          %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);

					t3=cvGetTickCount();
					dt=(double(t3-t1)/cvGetTickFrequency())/pow(10.0,6);
					//printf("histogram computing time: %1.5fsec.\n",dt);	

					//***  COMPLETED
					//***  HISTOGRAM IS IN "iblockhist"
					//********************************************************
		
					//*** detect objects
					tc1 = clock ();
					double dtsum=0.0;
					uint hist_size_x = iblockhist[0]->cols;
					uint hist_size_y = iblockhist[0]->rows;
					uint subwin_size_x = model.begin()->histblock_winsize_x;
					uint subwin_size_y = model.begin()->histblock_winsize_y;
					uint subwin_step_x = 1;
					uint subwin_step_y = 1;    
					uint top, left;
					
					if (hist_size_x>=subwin_size_x && hist_size_y>=subwin_size_y){
						uint left_max = hist_size_x - subwin_size_x;     assert( left_max > 0 );
						uint top_max  = hist_size_y - subwin_size_y;     assert( top_max > 0 );
						ClearModelCache(model);
						//*** classify each subwindow in the image
						for(left = 0; left < left_max; left += subwin_step_x){
							for(top = 0; top < top_max; top += subwin_step_y){
								t1=cvGetTickCount();
								SubWindowInfo result = adacascadeclass(model, iblockhist,
																	nc, left, top, 
																	subwin_size_x, subwin_size_y);
								if (result.passed){ // stage>=stage_min){
									result.x = (uint)((double)(result.x+1)*dx/scf);
									result.y = (uint)((double)(result.y+1)*dy/scf);
									result.w *= dx/scf;
									result.h *= dy/scf;
									subwindows.push_back(result);
								}
								t2=cvGetTickCount();
								dt=(double(t2-t1)/cvGetTickFrequency())/pow(10.0,6);
								dtsum+=dt;
							}
						} // iteration over all positions
						//printf("Detection time: %1.3fsec.\n",dtsum);
						t02=cvGetTickCount();
						dt=(double(t02-t01)/cvGetTickFrequency())/pow(10.0,6);
						//printf("Total detection time %1.3f\n",dt);
					}
					tc2 = clock ();
					totdettime+=(double)(tc2-tc1)/(double)CLK_TCK;
					//printf("  Detection time:                  %1.6fsec.\n",(double)(tc2-tc1)/(double)CLK_TCK);

					// */
					for (int i=0; i<nc; i++) cvReleaseMat(&iblockhist[i]);
					delete[] *iblockhist;
				}
			} // iteration over block sizes
			
			cvReleaseMat(&srcimg);
			cvReleaseMat(&srcimg32f);
			cvReleaseMat(&timg);
			cvReleaseMat(&wimg);
		}
	} // iterarion over scales
	
	//printf(" -> total processing time: %1.3f;\n",totproctime);
	//printf("    total detection time:  %1.3f\n",totdettime);

	cvReleaseMat(&srcgr);

	//*** pruning
	SubWindows prunedsubwindows;
    PruneSubWindows(subwindows, 0.3, 1, prunedsubwindows);

    return prunedsubwindows;
}

void ClearModelCache(Model const & model)
{
    for (Model::const_iterator it = model.begin(),
                              end = model.end();
                              it != end; ++it)
    {
        for (StageRules::const_iterator rit = it->rules.begin(),
                                       rend = it->rules.end();
                                       rit != rend; ++rit)
        {
            rit->clear_cache();
        }
    }
}

