#include <getopt.h>

#include "common.h"
#include "adacascadeclass.h"
#include "data.h"
#include "detectplate.h"
#include "dump.h"
#include "integralhistogram.h"
#include "prune_subwindows.h"
#include "profiling.h"

// #include <unistd.h>
#include <stdio.h>

#include <string>


/**
 * read the image data from an xml file
 */
CvMat * ReadFileImageXml(std::string const & image_xml_file,
                      std::string const & image_xml_tag);
CvMat * ReadFileImage(std::string const & image_file);



using namespace std;

int main(unsigned int argc, char ** argv)
{
    Tick("start");
    string basepath = "../..";
    string image_file       = "images/JPGImages/numberplates_0787.JPG";
    string image_xml_file   = "temp/tmpimg.xml";
    string image_xml_tag    = "srcimg";
    string texton_xml_file  = "data/detection/textons/textons_numberplate_set1_tnum12_scn2_ftrn5.xml";
    string inthist_xml_file = "temp/tmpinthistblock.xml";
    string model_xml_file   = "data/model.xml";

    /**
     * Command line arguments and options
     */
    char const * options_string = "b:x:n:t:o:m:i:";
    static struct option long_options[] =
    {
        {"basepath",    required_argument, 0, 'b'},
        {"image",       required_argument, 0, 'i'},
        {"imagexml",    required_argument, 0, 'x'},
        {"imagetag",    required_argument, 0, 'n'},
        {"parameters",  required_argument, 0, 't'},
        {"hist",        required_argument, 0, 'o'},
        {"model",       required_argument, 0, 'm'},
        {0, 0, 0, 0}
    };

    while(true)
    {
        int option_index = 0;
        
        int c = getopt_long (argc, argv, options_string,
                            long_options, &option_index);
        if (c == -1)
            break;

        switch(c)
        {
        case 'b':
            basepath = optarg;
            break;
        case 'i':
            image_file = optarg;
            break;
        case 'x':
            image_xml_file = optarg;
            break;
        case 'n':
            image_xml_tag = optarg;
            break;
        case 't':
            texton_xml_file = optarg;
            break;
        case 'o':
            inthist_xml_file = optarg;
            break;
        case 'm':
            model_xml_file = optarg;
            break;
        default:
            abort();
        }
    };

    // any remaining arguments
    if (optind < argc)
    {
        cerr << "Unexpected argument \"" << argv[optind] << "\"\n";
        return 1;
    }

    if (!basepath.empty())
    {
        // add the directory separator if needed
        char const back = basepath[basepath.size() - 1];
        if (back != '/' && back != '\\')
            basepath += '/';

        // prepend the base path
        image_file.insert(0, basepath);
        image_xml_file.insert(0, basepath);
        texton_xml_file.insert(0, basepath);
        inthist_xml_file.insert(0, basepath);
        model_xml_file.insert(0, basepath);
    }

    cout << "inthistblock:\n";
    cout << "  Input image :            " << image_file << "\n";
    cout << "  Input image (xml) :      " << image_xml_file << "\n";
    cout << "  Input image name :       " << image_xml_tag << "\n";
    cout << "  Input textons (xml) :    " << texton_xml_file << "\n";
    cout << "  Output histogram (xml) : " << inthist_xml_file << "\n";
    cout << "  Model file (xml) :       " << model_xml_file << "\n";

    Tick("Command line switches");

    Model model;
    xmlloadmodel( model_xml_file, model );

    Tick("Model loading");

    // read an image file
//    CvMat * source_image = ReadFileImageXml(image_xml_file, image_xml_tag);
    CvMat * source_image = ReadFileImage(image_file);
    Tick("image data reading");

    // read textons info
    pair< pair<CvMat*,CvMat*>, CvMat*> texton_info = ReadTextons(texton_xml_file);
    CvMat * cluster_means       = texton_info.first.first;
    CvMat * cov_matrix          = texton_info.first.second;
    CvMat * cluster_means_norm  = texton_info.second;
    Tick("Texton data reading");

    SubWindows subwindows =
            DetectPlate( source_image, model, cluster_means, cov_matrix, cluster_means_norm );

    DumpResultingWindows( basepath + "dump/subwindows.xml", subwindows );

    cvReleaseMat( &cluster_means );
    cvReleaseMat( &cov_matrix );
    cvReleaseMat( &cluster_means_norm );
    Tick("return 0");
    return 0;
}

CvMat * ReadFileImageXml(string const & image_xml_file, string const & image_xml_tag)
{
    CvFileStorage * fsr = cvOpenFileStorage(image_xml_file.c_str(), NULL, CV_STORAGE_READ);
//    CvFileNode * fn = cvGetFileNode(fsr, NULL);
    //CvFileNode * mn1 = cvGetFileNodeByName(fsr,0,"img3part");
    //CvFileNode* mn1 = cvGetFileNode(fsr, 0, "srcimg");

    // we have to clone the matrice itherwise it will be deleted with the file storage
    CvMat * srcimg = cvCloneMat((CvMat*)(cvRead(fsr, cvGetFileNodeByName(fsr, NULL, image_xml_tag.c_str()), NULL)));
    cvReleaseFileStorage(&fsr);
    return srcimg;
}

CvMat * ReadFileImage(std::string const & image_file)
{
    IplImage * image = cvLoadImage( image_file.c_str(), 1 ); // read as a color image
    if (!image)
        return NULL;
    int x = image->width;
    int y = image->height;
    if (!(x * y))
        return NULL;

    CvMat * source_image = cvCreateMat( y, x, CV_32F );

//    cvCopy(image, source_image);
    unsigned char * ptr = &CV_IMAGE_ELEM( image, unsigned char, 0, 0 );
    assert( sizeof(float) == 4 );
    float * sptr = (float*)(CV_MAT_ELEM_PTR_FAST( *source_image, 0, 0, sizeof(float) ));
    for (int j = 0; j < y; ++j)
    {
        for (int i = 0; i < x; ++i)
        {
            float sum = (*ptr);
            sum += (*++ptr);
            sum += (*++ptr);
            ++ptr;

            *(sptr) = sum / 3;
            ++sptr;
        }
    }
    DumpMatrix( "../../dump/src_32.xml", source_image );
    cvReleaseImage( &image );
    return source_image;
}
